;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "wfnames" "20260706.903"
  "Edit filenames."
  '((emacs "24.4"))
  :url "https://github.com/thierryvolpiatto/wfnames"
  :commit "d8839fa42a24f7c781cd2d8c3f40eda31faa19be"
  :revdesc "d8839fa42a24"
  :keywords '("wfnames" "convenience" "files" "editing" "helm")
  :authors '(("Thierry Volpiatto" . "thievol@posteo.net"))
  :maintainers '(("Thierry Volpiatto" . "thievol@posteo.net")))
